from pathlib import Path
from textwrap import dedent
from typing import Union

import pytest
from rich.console import Console
from rich.padding import Padding

from pytest_pointers.utils import FuncFinder

CACHE_TARGETS = "pointers/targets"
CACHE_ALL_FUNC = "pointers/funcs"


def pytest_addoption(parser):
    group = parser.getgroup("pointers")
    group.addoption(
        "--pointers-report",
        action="store_true",
        dest="pointers_report",
        default=False,
        help="Show report in console",
    )
    group.addoption(
        "--pointers-collect",
        action="store_true",
        dest="pointers_collect",
        default=False,
        help="Gather targets and tests for them",
    )


def pytest_configure(config):
    config.addinivalue_line("markers", "pointer(element): Define a tested element.")


def pytest_sessionstart(session: pytest.Session):
    if session.config.option.pointers_collect or session.config.option.pointers_report:
        session.config.cache.set(CACHE_TARGETS, {})


@pytest.fixture(scope="function", autouse=True)
def _pointer_marker(request):
    if (
        not request.config.option.pointers_collect
        and not request.config.option.pointers_report
    ):
        return

    pointers = request.config.cache.get(CACHE_TARGETS, {})
    marker = request.node.get_closest_marker("pointer")
    if marker:
        target_full_name = None
        if "target" in marker.kwargs:
            # support for regular pointers
            # @pytest.mark.pointer(target=Some.for_test)
            target = marker.kwargs.get("target", None)
            target_full_name = f"{target.__module__}.{target.__qualname__}"
        elif len(marker.args) == 2:
            # support for property pointers
            # @pytest.mark.pointer(Some, 'for_test')
            target_parent, target_member = marker.args
            target_full_name = f"{target_parent.__module__}.{target_parent.__qualname__}.{target_member}"

        if target_full_name not in pointers:
            pointers[target_full_name] = []

        pointers[target_full_name].append(request.node.nodeid)
        request.config.cache.set(CACHE_TARGETS, pointers)


def pytest_sessionfinish(
    session: pytest.Session, exitstatus: Union[int, pytest.ExitCode]
):
    if session.config.option.pointers_report:
        console = Console()
        console.print("")

        pointers = session.config.cache.get(CACHE_TARGETS, {})

        start_dir = Path(session.startdir)  # noqa
        funcs = FuncFinder(start_dir)

        def report_line(f):
            test_count = len(pointers.get(f, []))

            if test_count > 1:
                color = "green"
            elif test_count == 1:
                color = "blue"
            else:
                color = "red"

            test_count_str = f"{test_count: <2}"
            return f"[{color}]{test_count_str:·<5}[/{color}] {f}"

        report_lines = "\n".join([report_line(f) for f in funcs])
        report = dedent(
            f"""
        [bold]List of functions in project and the number of tests for them[/bold]

        \n{report_lines}
        """
        )

        test = Padding(report, (2, 4), expand=False)
        console.print(test)



from pathlib import Path
from textwrap import dedent

import libcst as cst
import pytest

from pytest_pointers.tests.mock_structure import for_func_finder
from pytest_pointers.utils import FuncFinder


@pytest.mark.pointer(target=FuncFinder.get_methods_qual_names)
class TestFuncFinderGetNodeQualName:
    @pytest.mark.parametrize(
        ("module", "expect"),
        [
            pytest.param(
                dedent(
                    """\
                    def for_test():
                        pass
                    """
                ),
                ["for_test"],
                id="Simple function",
            ),
            pytest.param(
                dedent(
                    """\
                    class Some:
                        def __init__(self):
                            pass

                        def for_test(self):
                            pass
                    """
                ),
                ["Some.__init__", "Some.for_test"],
                id="Multiple methods",
            ),
            pytest.param(
                dedent(
                    """\
                        class Some:
                            async def for_test(self):
                                pass
                        """
                ),
                ["Some.for_test"],
                id="Async methods",
            ),
            pytest.param(
                dedent(
                    """\
                        def some_decor(f):
                            def wrapper(*args, **kwargs):
                                f(*args, **kwargs)
                            return wrapper
                            
                        class Some:
                            @some_decor
                            def for_test(self):
                                pass
                        """
                ),
                ["some_decor", "Some.for_test"],
                id="Support for decorators",
            ),
            pytest.param(
                dedent(
                    """\
                        class Some:
                            def __init__(self):
                                pass

                            def for_test(self):
                                pass


                        class Another:
                            def __init__(self):
                                pass

                            def for_test(self):
                                pass
                        """
                ),
                [
                    "Some.__init__",
                    "Some.for_test",
                    "Another.__init__",
                    "Another.for_test",
                ],
                id="Multiple classes with same methods",
            ),
            pytest.param(
                dedent(
                    """\
                    class Some:
                        def for_test(self):  # notest:
                            pass
                    """
                ),
                [],
                id="Ignored method must be omitted",
            ),
        ],
    )
    def test__func_called__result_returned(self, module, expect):
        tree = cst.parse_module(module)

        result = list(FuncFinder.get_methods_qual_names(tree))

        assert result == expect


@pytest.mark.pointer(target=FuncFinder.__iter__)
class TestFuncFinderIter:
    def test__func_called__result_returned(
        self,
    ):
        funcs = FuncFinder(Path(for_func_finder.__file__).parent)

        assert set(funcs) == {
            "pytest_pointers.tests.mock_structure.for_func_finder.Some.__init__",
            "pytest_pointers.tests.mock_structure.for_func_finder.Some.for_test",
            "pytest_pointers.tests.mock_structure.for_func_finder.Another.__init__",
            "pytest_pointers.tests.mock_structure.for_func_finder.for_test",
            "pytest_pointers.tests.mock_structure.for_func_finder.Another.for_test",
        }


@pytest.mark.pointer(target=FuncFinder.get_py_files)
class TestFuncFinderGetPyFiles:
    @pytest.mark.parametrize(
        ("files", "files_ignored"),
        [
            (
                [
                    "dir1/module.py",
                    "dir1/module2.py",
                ],
                [
                    ".venv/pack1/module.py",
                    ".venv/pack2/module.py",
                ],
            ),
            (
                [
                    "dir1/module.py",
                    "dir2/pack1/m.py",
                    "dir2/pack1/m2.py",
                ],
                [
                    ".venv/pack1/module.py",
                ],
            ),
            (
                [
                    "dir1/module.py",
                    "dir2/pack1/m.py",
                    "dir2/pack1/m2.py",
                ],
                [
                    "dir1/tests/test_module.py",
                ],
            ),
        ],
    )
    def test__func_called__result_returned(self, tmp_path, files, files_ignored):
        files_path = [tmp_path / Path(f) for f in files]
        files_ignored_path = [tmp_path / Path(f) for f in files_ignored]

        for f in files_path + files_ignored_path:
            f.parent.mkdir(parents=True, exist_ok=True)
            f.touch()

        result = FuncFinder(start_dir=tmp_path).get_py_files()
        assert len(result - set(files_path)) == 0



def for_test():
    pass


class Some:
    def __init__(self):
        pass

    def for_test(self):
        pass


class Another:
    def __init__(self):
        pass

    def for_test(self):
        pass



import json

import pytest

from pytest_pointers.plugin import _pointer_marker

pytestmark = [pytest.mark.unit]


@pytest.mark.pointer(target=_pointer_marker)
class TestPointerMarkerFixture:
    @staticmethod
    def get_params():
        content = """\
            import pytest

            class Some:
                def for_test(self):
                    return 1

            @pytest.mark.pointer(target=Some.for_test)
            def test_for_test():
                result = Some().for_test()
                assert result == 1
        """
        expect = {
            "test__regular_flow__pointers_collected.Some.for_test": [
                "test__regular_flow__pointers_collected.py::test_for_test"
            ]
        }
        yield pytest.param(content, expect, id="One test of method")

        # ----8<----------------------------------------------------------------------

        content = """\
            import pytest

            class Some:
                def for_test1(self):
                    return 1

                def for_test2(self):
                    return 1

            @pytest.mark.pointer(target=Some.for_test1)
            def test_for_test1():
                result = Some().for_test1()
                assert result == 1

            @pytest.mark.pointer(target=Some.for_test2)
            def test_for_test2():
                result = Some().for_test2()
                assert result == 1
        """
        expect = {
            "test__regular_flow__pointers_collected.Some.for_test1": [
                "test__regular_flow__pointers_collected.py::test_for_test1"
            ],
            "test__regular_flow__pointers_collected.Some.for_test2": [
                "test__regular_flow__pointers_collected.py::test_for_test2"
            ],
        }
        yield pytest.param(content, expect, id="Two tests for two methods")

        # ----8<----------------------------------------------------------------------

        content = """\
            import pytest

            class Some:
                def for_test(self):
                    return 1


            @pytest.mark.pointer(target=Some.for_test)
            class TestSome:
                def test_for_test1(self):
                    result = Some().for_test()
                    assert result == 1

                def test_for_test2(self):
                    result = Some().for_test()
                    assert result == 1
        """
        expect = {
            "test__regular_flow__pointers_collected.Some.for_test": [
                "test__regular_flow__pointers_collected.py::TestSome::test_for_test1",
                "test__regular_flow__pointers_collected.py::TestSome::test_for_test2",
            ]
        }
        yield pytest.param(content, expect, id="Two test for one method")

        # ----8<----------------------------------------------------------------------

        content = """\
            import pytest

            class Some:
                @property
                def for_test(self):
                    return 1


            @pytest.mark.pointer(Some, 'for_test')
            class TestSome:
                def test_for_test(self):
                    result = Some().for_test
                    assert result == 1
        """
        expect = {
            "test__regular_flow__pointers_collected.Some.for_test": [
                "test__regular_flow__pointers_collected.py::TestSome::test_for_test"
            ]
        }
        yield pytest.param(content, expect, id="Property as a target")

        # ----8<----------------------------------------------------------------------

        content = """\
            import pytest

            class Some:
                @classmethod
                def for_test(cls):
                    return 1
            
            
            @pytest.mark.pointer(target=Some.for_test)
            class TestSome:
                def test_for_test(self):
                    result = Some.for_test()
                    assert result == 1
        """
        expect = {
            "test__regular_flow__pointers_collected.Some.for_test": [
                "test__regular_flow__pointers_collected.py::TestSome::test_for_test"
            ]
        }
        yield pytest.param(content, expect, id="Classmethod as a target")

    @pytest.mark.parametrize(("content", "expect"), get_params.__func__())  # noqa
    def test__regular_flow__pointers_collected(self, testdir, content, expect):
        """
        Make sure that pointer marker is working properly
        """
        testdir.makepyfile(content)

        result = testdir.runpytest("--pointers-collect")
        assert result.ret == 0

        cache_file = testdir.tmpdir / ".pytest_cache" / "v" / "pointers" / "targets"
        assert cache_file.exists()
        with open(cache_file, "r") as f:
            targets = json.loads(f.read())

        assert targets == expect



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pytest_pointers',
 'pytest_pointers.tests',
 'pytest_pointers.tests.mock_structure']

package_data = \
{'': ['*']}

install_requires = \
['libcst>=0.3.15,<0.4.0', 'rich']

entry_points = \
{'console_scripts': ['format = poetry_deps_formatter:formatter'],
 'pytest11': ['plugin = pytest_pointers.plugin']}

setup_kwargs = {
    'name': 'pytest-pointers',
    'version': '0.2.7',
    'description': 'Pytest plugin to define functions you test with special marks for better navigation and reports',
    'long_description': '## Pytest Plugin to show real test coverage\n\n### TLTR\n\nWith this plugin you will be able to see all methods in project with the number of their tests\n\nJust call:\n```\npytest --pointers-report -v\n```\n\n![](https://jaklimoff-misc.s3.eu-central-1.amazonaws.com/pytest-pointers/example_output.jpg)\n',
    'author': 'Jack Klimov',
    'author_email': 'jaklimoff@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/jaklimoff/pytest-pointers',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



import sys
from dataclasses import dataclass
from pathlib import Path
from typing import Set

import libcst as cst
from libcst.metadata import QualifiedNameProvider, ParentNodeProvider


class MethodQualNamesCollector(cst.CSTVisitor):
    METADATA_DEPENDENCIES = (QualifiedNameProvider, ParentNodeProvider)

    def __init__(self):
        self.found = []
        super().__init__()

    def visit_FunctionDef(self, node: cst.FunctionDef):
        header = getattr(node.body, "header", None)
        excluded = (
            header is not None
            and header.comment
            and header.comment.value.find("notest:") > -1
        )

        if not excluded:
            # TODO: Find better way to remove locals
            qual_names = self.get_metadata(QualifiedNameProvider, node)
            for qn in qual_names:
                from_local = qn.name.find("<locals>") > -1
                if not from_local:
                    self.found.append(qn.name)


@dataclass
class FuncFinder:
    start_dir: Path

    @classmethod
    def get_methods_qual_names(cls, node: cst.Module):
        collector = MethodQualNamesCollector()
        cst.MetadataWrapper(node).visit(collector)
        for method_name in collector.found:
            yield method_name

    def get_py_files(self) -> Set[Path]:
        py_files = [p for p in self.start_dir.glob("**/*.py")]
        py_files = set(py_files)

        for pattern in [".venv/**/*.py", "venv/**/*.py", "tests/**/*.py"]:
            py_files = py_files - set(self.start_dir.rglob(pattern))

        return py_files

    def __iter__(self):
        source_paths = [Path(p) for p in sys.path]

        py_files = self.get_py_files()

        for p in py_files:
            with open(p, "r") as f:
                tree = cst.parse_module(f.read())

            source = min(set(p.parents) & set(source_paths))
            abs_import = p.parts[len(source.parts) : -1] + (p.stem,)

            for fun_name in self.get_methods_qual_names(tree):
                yield ".".join(abs_import + (fun_name,))
